import daj.*;
import java.util.*;

class Group
{
	private int Coordinator;
	private int Counter;

	public Group (int coordinator, int counter)
	{
		Coordinator = coordinator;
		Counter = counter;
	}

	public boolean equals (Object obj)
	{
		if (obj instanceof Group)
		{
			Group group = (Group)obj;
			return (Coordinator == group.Coordinator && Counter == group.Counter);
		}
		return false;
	}

	public String toString ()
	{
		return "[" + String.valueOf (Coordinator) + ":" + String.valueOf (Counter) + "]";
	}
};

/**
	The main program to be executed on each machine
*/
class Processor extends Program
{
	public static final int DELAY = 3;
	public static final int TIMEOUT = 20;

	public static final int NORMAL = 0;
	public static final int DOWN = 1;
	public static final int ELECTION = 2;
	public static final int REORGANIZATION = 3;

	/**
		The index of this node
	*/
	public int index = 0;

	/**
		The index of the coordinator of the group
	*/
	public int Coordinator = 0;

	/**
		The group to which this node belongs
	*/
	public Group Group = null;

	/**
		The counter for generating unique groups
	*/
	public int Counter = -1;

	/**
		The state of the processor
	*/
	public int State = NORMAL;

	/**
		The set of Up processors
	*/
	public BitSet Up = new BitSet ();

	protected Node selfNode = null;
	protected MainThread mainThread = new MainThread (this);
	protected Node mainNode = null;
	protected InvitationThread invitationThread = new InvitationThread (this);
	protected Node invitationNode = null;
	protected ElectionMonitorThread electionMonitorThread = new ElectionMonitorThread (this);
	protected Node electionMonitorNode = null;

	public void main ()
	{
		Random random = new Random ();

		Recovery ();
		while (true)
		{
			int received = in ().select (Math.abs (random.nextInt ()) % 5);
			if (received == -1)
				continue;

			// Simulate failure
			if (Math.abs (random.nextInt ()) % 20 == 0)
			{
				State = DOWN;
				sleep (10 + Math.abs (random.nextInt ()) % 20);
				Recovery ();
				continue;
			}

			Message msg = in (received).receive ();
			if (msg == null)
				return;
			if (received >= getNeighborsCount ())
			{
				// Simulate a delay in sending
				if (Math.abs (random.nextInt ()) % 10 == 0)
					sleep (Math.abs (random.nextInt ()) % DELAY);

				// Internal message;unpack it
				InternalMessage intMsg = (InternalMessage)msg;
				if (intMsg.channel < 0)
				{
					// Broadcast it
					for (int channel = 0;channel < getNeighborsCount ();channel++)
						out (channel).send (intMsg.outerMessage);
				}
				else
				{
					out (intMsg.channel).send (intMsg.outerMessage);
				}
			}
			else
			{
				// Normal message;pack it and choose the right destination
				InternalMessage intMsg = new InternalMessage (received, msg);
				if (msg instanceof AreYouCoordAnswer || msg instanceof ReadyAnswer)
				{
					out (getMainChannel ()).send (intMsg);
					continue;
				}
				if (msg instanceof InvitationMsg || msg instanceof AcceptAnswer)
				{
					out (getInvitationChannel ()).send (intMsg);
					continue;
				}
				out (getElectionMonitorChannel ()).send (intMsg);
			}
		}
	}

	public Node OnNode (Application app, int index, int x, int y)
	{
		this.index = index;
		Group = new Group (index, Counter);

		mainNode = app.node (mainThread, "", x, y - 20);
		invitationNode = app.node (invitationThread, "", x - 17, y + 10);
		electionMonitorNode = app.node (electionMonitorThread, "", x + 17, y + 10);
		selfNode = app.node (this, String.valueOf (index), x, y);

		return selfNode;
	}

	public void OnLink (Application app)
	{
		app.link (selfNode, mainNode);
		app.link (mainNode, selfNode);

		app.link (selfNode, invitationNode);
		app.link (invitationNode, selfNode);

		app.link (selfNode, electionMonitorNode);
		app.link (electionMonitorNode, selfNode);
		}

	public int getNeighborsCount ()
	{
		if (in () != null)
		{
			int count = in ().getSize ();
			if (mainNode != null)
				count--;
			if (invitationNode != null)
				count--;
			if (electionMonitorNode != null)
				count--;
			return count;
		}
		else
			return 0;
	}

	public int getMainChannel ()
	{
		return getNeighborsCount ();
	}

	public int getInvitationChannel ()
	{
		return getNeighborsCount () + 1;
	}

	public int getElectionMonitorChannel ()
	{
		return getNeighborsCount () + 2;
	}

	/**
		Recovers after failure
	*/
	public void Recovery ()
	{
		State = ELECTION;
		Group = new Group (index, ++Counter);
		Coordinator = index;
		Up = new BitSet ();
		State = Processor.REORGANIZATION;
		State = Processor.NORMAL;
	}

	public String getText ()
	{
		if (State == DOWN)
			return "Failed";

		String s = "";
		s += "Group: " + this.Group.toString () + "\n";

		String states[] = {"Normal", "Down", "Election", "Reorganization"};
		s += "State: " + states[State] + "\n";

		return s;
	}
}

